﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using WorkflowLibrary;
using System.Workflow.Runtime;
using System.Workflow.Activities;
using System.Workflow.ComponentModel;
using System.Threading;
using StateMachineTracking;
using System.Collections.ObjectModel;
using WorkflowDesignerControl;
using System.IO;
using System.Drawing.Imaging;
using System.Collections;

namespace WorkflowApplication
{
    public partial class WorkflowApplication : Form, ITaskService
    {
         
        private WorkflowRuntime workflowRuntime;
        private WorkflowRuntime WFRuntime
        {
            get
            {
                if (workflowRuntime == null)
                    return startWorkflowRuntime();
                
                if(!workflowRuntime.IsStarted)
                    workflowRuntime.StartRuntime();

                return workflowRuntime;
            }
        }

        private string selectedInstanceID; 
        private WorkflowInstance WFInstance
        {
            get
            {
                return InstanceHolder.Get(selectedInstanceID).WFInstance;
            }
        }

        private StateMachineTrackingService stateMachineTrackingService;

        public Type WFType
        {
            get 
            {
                InstanceHolder instanceHolder = InstanceHolder.Get(selectedInstanceID);

                if (instanceHolder.WFType == WorkflowType.Sequential)
                    return typeof(SequentialWorkflow);
                else
                    if (instanceHolder.WFType == WorkflowType.Statemachine)
                        return typeof(StateMachineWorkflow);
                    else
                        throw new Exception("Unsupported workflow type");
            }
        }

        private string currentActivityName = string.Empty;
        
        public string CurrentActivityName
        {
            get
            {
                return currentActivityName;                
            }
            set
            {
                currentActivityName = value;
            }
        }

        public WorkflowApplication()
        {
            InitializeComponent();
        }

        private void sequentialToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // стартира последователен поток
            WorkflowInstance sequentialInstance = WFRuntime.CreateWorkflow(typeof(SequentialWorkflow));
            sequentialInstance.Start();
            selectedInstanceID = sequentialInstance.InstanceId.ToString();
            InstanceHolder.Add(WorkflowType.Sequential, sequentialInstance);
            CurrentActivityName = string.Empty;            

            CreateTaskForm form = new CreateTaskForm();
            DialogResult result = form.ShowDialog();
            if(result == DialogResult.OK)
                if (TaskCreated != null)
                {
                    this.TaskCreated(null, new TaskEventArgs(WFInstance.InstanceId, form.Severity));
                    AddListViewItem(WFInstance.InstanceId, form.GetTask);
                }

        }


        #region ITaskService Members

        public event EventHandler<TaskEventArgs> TaskCreated;

        public event EventHandler<TaskEventArgs> TaskRunning;

        public event EventHandler<TaskEventArgs> TaskQA;

        public event EventHandler<TaskEventArgs> TaskTerminated;

        public event EventHandler<TaskEventArgs> TaskCompleted;

        public event EventHandler<TaskEventArgs> CodeReviewed;

        #endregion

        private delegate void SendMessageDelegate(string message);

        private delegate void AddStatusDelegate(string message);

        private delegate void UpdateListViewItemDelegate(Guid workflowInstanceID, Task task);

        private delegate void UpdateListViewItemStatusDelegate(Guid workflowInstanceID, TaskStatus taskStatus);

        private delegate void RefreshUIDelegate();

        private void AddStatusLabel(string message)
        {
            if (this.InvokeRequired)
            {
                AddStatusDelegate addStatusDelegate = new AddStatusDelegate(this.AddStatusLabel);
                object[] args = new object[1] { message };
                this.Invoke(addStatusDelegate, args);
            }
            else
            {
                statusLabel.Text = message;
            }

        }

        public void SendMessage(string message)
        {
            if (logListView.InvokeRequired)
            {
                // Този код се изпълнява на различна нишка от тази която е създала
                // списъка, така че е необходимо използване на Invoke метода
                // Създава инстанция на делегата за извикване на този метод
                SendMessageDelegate sendMessageDelegate = new SendMessageDelegate(this.SendMessage);

                // Създава масив от параметри за този метод
                object[] args = new object[1] { message };

                // Извиква метода върху нишката на потребителския интерфейс
                this.Invoke(sendMessageDelegate, args);
            }
            else
            {
                //Добавя съобщението към лога
                logListView.Items.Add(message);
            }

        }

        void workflowRuntime_WorkflowCompleted(object sender, WorkflowCompletedEventArgs e)
        {
            SendMessage("");
            AddStatusLabel("Workflow completed");
            UpdateListViewItem(WFInstance.InstanceId, TaskStatus.Completed);
            DisableDropDownButtons();
            //WFRuntime.StopRuntime();
         }

        void workflowRuntime_WorkflowCreated(object sender, WorkflowEventArgs e)
        {
            AddStatusLabel("Workflow created");
        }


        private WorkflowRuntime startWorkflowRuntime()
        {
            workflowRuntime = new WorkflowRuntime();

            ExternalDataExchangeService dataService = new ExternalDataExchangeService();
            workflowRuntime.AddService(dataService);
            dataService.AddService(this);

            stateMachineTrackingService = new StateMachineTrackingService(workflowRuntime);

            workflowRuntime.WorkflowCreated += new EventHandler<WorkflowEventArgs>(workflowRuntime_WorkflowCreated);
            workflowRuntime.WorkflowStarted += new EventHandler<WorkflowEventArgs>(workflowRuntime_WorkflowStarted);
            workflowRuntime.WorkflowCompleted += new EventHandler<WorkflowCompletedEventArgs>(workflowRuntime_WorkflowCompleted);
            
            workflowRuntime.StartRuntime();
            
            return workflowRuntime;
        }

        void workflowRuntime_WorkflowStarted(object sender, WorkflowEventArgs e)
        {
            AddStatusLabel("Workflow started");
        }


        private void codeReviewToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (CodeReviewed != null)
            {
                CodeReviewed(null, new TaskEventArgs(WFInstance.InstanceId, "Code reviewed"));
            }
        }

        private void completeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (TaskCompleted != null)
            {
                TaskCompleted(null, new TaskEventArgs(WFInstance.InstanceId, "Task completed"));
            }
        }

        private ListViewItem AddListViewItem(Guid workflowInstanceID, Task task)
        {
            // Прибавя нов елемент към ListView контролата
            ListViewItem listViewItem =
                this.taskStatusListView.Items.Add(workflowInstanceID.ToString(),
                workflowInstanceID.ToString(), "");

            // Прибавя колоните име, статус, трудност
            listViewItem.SubItems.Add(task.Name);
            listViewItem.SubItems.Add(task.Status.ToString());
            listViewItem.SubItems.Add(task.Severity.ToString());
            listViewItem.Tag = workflowInstanceID.ToString();

            // Избира новия ListViewItem
            listViewItem.Selected = true;

            // Връща ListViewItem
            return listViewItem;
        }

        private void UpdateListViewItem(Guid workflowInstanceID, Task task)
        {

            if (this.taskStatusListView.InvokeRequired)
            {
                UpdateListViewItemDelegate updateListViewItemDelegate =
                    new UpdateListViewItemDelegate(this.UpdateListViewItem);

                object[] args = new object[2] { workflowInstanceID, task};

                this.taskStatusListView.Invoke(updateListViewItemDelegate, args);
            }
            else
            {
                // взема ListViewItem за дадената инстанция
                ListViewItem listViewItem = taskStatusListView.Items[workflowInstanceID.ToString()];

                if (listViewItem == null)
                {
                    return;
                }

                // Update the Workflow State & Status column values
                listViewItem.SubItems[1].Text = task.Name;
                listViewItem.SubItems[2].Text = task.Status.ToString();
                listViewItem.SubItems[3].Text = task.Severity.ToString();
            }
        }

        private void UpdateListViewItem(Guid workflowInstanceID, TaskStatus taskStatus)
        {

            if (this.taskStatusListView.InvokeRequired)
            {
                UpdateListViewItemStatusDelegate updateListViewItemDelegate =
                    new UpdateListViewItemStatusDelegate(this.UpdateListViewItem);

                object[] args = new object[2] { workflowInstanceID, taskStatus };

                this.taskStatusListView.Invoke(updateListViewItemDelegate, args);
            }
            else
            {
                // взема ListViewItem за дадената инстанция
                ListViewItem listViewItem = taskStatusListView.Items[workflowInstanceID.ToString()];

                if (listViewItem == null)
                {
                    return;
                }

                // Update the Workflow State & Status column values
                listViewItem.SubItems[2].Text = taskStatus.ToString();
            }
        }

        private void qAToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (TaskQA != null)
            {
                TaskQA(null, new TaskEventArgs(WFInstance.InstanceId, "Task QA"));
            }
        }

        private void runningToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (TaskRunning != null)
            {
                TaskRunning(null, new TaskEventArgs(WFInstance.InstanceId, "Task running"));
            }
        }

        private void completedToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (TaskCompleted != null)
            {
                TaskCompleted(null, new TaskEventArgs(WFInstance.InstanceId, "Task completed"));
            }
        }

        private void terminatedToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (TaskTerminated != null)
            {
                TaskTerminated(null, new TaskEventArgs(WFInstance.InstanceId, "Task terminated"));
            }
        }

        private void stateMachineToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // Стартира поток машина на състоянието
            if (stateMachineTrackingService == null)
                startWorkflowRuntime();

            // стартира поток машина на състоянието
            StateMachineInstance stateMachineInstance = stateMachineTrackingService.RegisterInstance(
                typeof(StateMachineWorkflow), Guid.NewGuid());
            stateMachineInstance.StateChanged += new EventHandler<ActivityEventArgs>(stateMachineInstance_StateChanged);
            stateMachineInstance.StartWorkflow();
            selectedInstanceID = stateMachineInstance.InstanceId.ToString();
            InstanceHolder.Add(WorkflowType.Statemachine, stateMachineInstance);
            
            
            // създава нова задача
            CreateTaskForm form = new CreateTaskForm();
            DialogResult result = form.ShowDialog();
            if (result == DialogResult.OK)
                if (TaskCreated != null)
                {
                    this.TaskCreated(null, new TaskEventArgs(WFInstance.InstanceId, form.Severity));
                    AddListViewItem(WFInstance.InstanceId, form.GetTask);
                }
        }

        void stateMachineInstance_StateChanged(object sender, ActivityEventArgs e)
        {
            StateMachineInstance statemachineInstance = InstanceHolder.Get(selectedInstanceID).SMInstance;
            if (e.QualifiedId != statemachineInstance.InitialState &&
                e.QualifiedId != statemachineInstance.CompletedState)
            {
                RefreshUI();
            }
        }

        private void RefreshUI()
        {
            if (this.InvokeRequired)
            {
                RefreshUIDelegate refreshUIDelegate = new RefreshUIDelegate(this.RefreshUI);
                this.Invoke(refreshUIDelegate);
            }
            else
            {
                if (InstanceHolder.Get(selectedInstanceID).WFType == WorkflowType.Sequential)
                    return;

                string workflowStatus = taskStatusListView.SelectedItems[0].SubItems[2].Text;
                if (workflowStatus.Equals("Completed"))
                    return;

                StateMachineWorkflowInstance instance = new StateMachineWorkflowInstance(WFRuntime, WFInstance.InstanceId);

                if (instance.CurrentState == null)
                    return;
                
                ReadOnlyCollection<string> transitions = instance.PossibleStateTransitions;

                DisableCommands();

                if (transitions.Contains("Running"))
                    runningToolStripMenuItem.Enabled = true;
                if (transitions.Contains("QA"))
                    qAToolStripMenuItem.Enabled = true;
                if (transitions.Contains("Completed"))
                    completedToolStripMenuItem.Enabled = true;
                if (transitions.Contains("Terminated"))
                    terminatedToolStripMenuItem.Enabled = true;

                CurrentActivityName = instance.CurrentStateName;
                UpdateListViewItem(instance.InstanceId, GetTaskStatus(CurrentActivityName));
            }
        }

        private TaskStatus GetTaskStatus(string statusName)
        {
            return (TaskStatus)Enum.Parse(typeof(TaskStatus), statusName);
        }
        private void DisableCommands()
        {
            runningToolStripMenuItem.Enabled = false;
            qAToolStripMenuItem.Enabled = false;
            completedToolStripMenuItem.Enabled = false;
            terminatedToolStripMenuItem.Enabled = false;
        }

        private void DisableDropDownButtons()
        {
            if (this.InvokeRequired)
            {
                RefreshUIDelegate refreshDelegate = new RefreshUIDelegate(DisableDropDownButtons);
                this.Invoke(refreshDelegate);
            }
            else
            {
                sequentialDropDownButton.Enabled = false;
                stateMachineDropDownButton.Enabled = false;
                wfImageToolStripButton.Enabled = false;
            }
        }

        private void designerToolStripButton_Click(object sender, EventArgs e)
        {
            DesignerForm designerForm = new DesignerForm();
            designerForm.Show();
        }

        private void wfImageToolStripButton_Click(object sender, EventArgs e)
        {
            OpenImage(0);
        }

        private void WorkflowApplication_Load(object sender, EventArgs e)
        {
            DisableDropDownButtons();
        }

        private void taskStatusListView_ItemSelectionChanged(object sender, ListViewItemSelectionChangedEventArgs e)
        {
            if (e.Item.Selected)
            {
                selectedInstanceID = taskStatusListView.SelectedItems[0].Text;
                sequentialDropDownButton.Enabled = (InstanceHolder.Get(selectedInstanceID).WFType == WorkflowType.Sequential);
                stateMachineDropDownButton.Enabled = (InstanceHolder.Get(selectedInstanceID).WFType == WorkflowType.Statemachine);
                wfImageToolStripButton.Enabled = true;
                RefreshUI();
            }

        }

        private void xamlToolStripButton_Click(object sender, EventArgs e)
        {
            OpenImage(1);
        }

        private void OpenImage(int activePage)
        {
            if (WFInstance == null)
                return;

            using (WorkflowDesignerControl.WorkflowDesignerControl designer = new WorkflowDesignerControl.WorkflowDesignerControl())
            {
                designer.WFType = WFType;
                using (MemoryStream ms = new MemoryStream())
                {
                    if (CurrentActivityName != String.Empty)
                        designer.HighlightActivity(CurrentActivityName);
                    designer.SaveWorkflowImage(ms, ImageFormat.Bmp);
                    Bitmap bitmap = new Bitmap(ms);
                    ImageForm imageForm = new ImageForm();
                    imageForm.WindowState = FormWindowState.Maximized;// Bounds = Screen.PrimaryScreen.Bounds;
                    imageForm.ImageTab.BackgroundImageLayout = ImageLayout.Zoom;
                    imageForm.ImageTab.BackgroundImage = bitmap;
                    imageForm.XAMLText.Text = LoadXAML();
                    imageForm.TabCtrl.SelectedIndex = activePage;
                    imageForm.Show();
                }
            }
        }

        private string LoadXAML()
        {
            StreamReader objReader = new StreamReader(WFType.Name + ".xoml");
            return objReader.ReadToEnd();
        }

    }

    
}