//--------------------------------------------------------------------------------
// This file is a "Sample" as from Windows Workflow Foundation
// Hands on Labs RC
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// This source code is intended only as a supplement to Microsoft
// Development Tools and/or on-line documentation.  See these other
// materials for detailed information regarding Microsoft code samples.
// 
// THIS CODE AND INFORMATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY
// KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//--------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.ComponentModel.Design;
using System.ComponentModel.Design.Serialization;
using System.Workflow.ComponentModel;
using System.Workflow.ComponentModel.Compiler;
using System.Workflow.Activities;
using System.Workflow.ComponentModel.Design;
using System.Runtime.InteropServices;
using System.Collections;
using System.Collections.ObjectModel;
using System.IO;
using System.Workflow.ComponentModel.Serialization;
using System.Xml;
using System.CodeDom.Compiler;
using System.Workflow.Runtime;
using System.Reflection;
using System.Threading;
using System.Drawing.Imaging;

namespace WorkflowDesignerControl
{
    public partial class WorkflowDesignerControl : UserControl, IDisposable, IServiceProvider
    {
        private WorkflowView workflowView;
        private DesignSurface designSurface;
        private WorkflowLoader loader;
        static Assembly inMemoryAssembly;
        private WorkflowRuntime workflowRuntime;
        SequentialWorkflowActivity workflow;

        public WorkflowDesignerControl()
        {
            InitializeComponent();

            Toolbox toolbox = new Toolbox(this);
            this.propertyGridSplitter.Panel1.Controls.Add(toolbox);
            toolbox.Dock = DockStyle.Fill;
            toolbox.BackColor = BackColor;
            toolbox.Font = WorkflowTheme.CurrentTheme.AmbientTheme.Font; 

            WorkflowTheme.CurrentTheme.ReadOnly = false;
            WorkflowTheme.CurrentTheme.AmbientTheme.ShowConfigErrors = false;
            WorkflowTheme.CurrentTheme.ReadOnly = true;

            this.propertyGrid.BackColor = BackColor;
            this.propertyGrid.Font = WorkflowTheme.CurrentTheme.AmbientTheme.Font;

            AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(CurrentDomain_AssemblyResolve);
        }

        static Assembly CurrentDomain_AssemblyResolve(object sender, ResolveEventArgs args)
        {
            if (inMemoryAssembly.FullName == args.Name)
            {
                return inMemoryAssembly;
            }
            else
            {
                return null;
            }
        }

        public string XomlFile
        {
            get
            {
                return this.loader.XomlFile;
            }
            set
            {
                this.loader.XomlFile = value;
            }
        }

        public string Xoml
        {
            get
            {
                string xoml = string.Empty;
                if (this.loader != null)
                {
                    try
                    {
                        this.loader.Flush();
                        xoml = this.loader.Xoml;
                    }
                    catch
                    {
                    }
                }
                return xoml;
            }

            set
            {
                try
                {
                    if (!String.IsNullOrEmpty(value))
                        LoadWorkflow(value);
                }
                catch
                {
                }
            }
        }

        public string WorkflowName
        {
            get
            {
                return this.workflow == null ? string.Empty : this.workflow.Name;
            }
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing && (components != null))
            {
                UnloadWorkflow();
                components.Dispose();
            }
            base.Dispose(disposing);
        }

        new public object GetService(Type serviceType)
        {
            return (this.workflowView != null) ? ((IServiceProvider)this.workflowView).GetService(serviceType) : null;
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);
            ShowDefaultWorkflow();
        }

        private void LoadWorkflow(string xoml)
        {
            SuspendLayout();

            DesignSurface designSurface = new DesignSurface();
            WorkflowLoader loader = new WorkflowLoader();
            loader.Xoml = xoml;
            designSurface.BeginLoad(loader);

            IDesignerHost designerHost = designSurface.GetService(typeof(IDesignerHost)) as IDesignerHost;
            if (designerHost != null && designerHost.RootComponent != null)
            {
                IRootDesigner rootDesigner = designerHost.GetDesigner(designerHost.RootComponent) as IRootDesigner;
                if (rootDesigner != null)
                {
                    UnloadWorkflow();

                    this.designSurface = designSurface;
                    this.loader = loader;
                    this.workflowView = rootDesigner.GetView(ViewTechnology.Default) as WorkflowView;
                    this.workflowViewSplitter.Panel1.Controls.Add(this.workflowView);
                    this.workflowView.Dock = DockStyle.Fill;
                    this.workflowView.TabIndex = 1;
                    this.workflowView.TabStop = true;
                    this.workflowView.HScrollBar.TabStop = false;
                    this.workflowView.VScrollBar.TabStop = false;
                    this.workflowView.Focus();

                    ISelectionService selectionService = GetService(typeof(ISelectionService)) as ISelectionService;

                    if (selectionService != null)
                    {
                        selectionService.SelectionChanged += new EventHandler(OnSelectionChanged);
                    }
                }
            }

            ResumeLayout(true);
        }

        private void UnloadWorkflow()
        {
            IDesignerHost designerHost = GetService(typeof(IDesignerHost)) as IDesignerHost;
            if (designerHost != null && designerHost.Container.Components.Count > 0)
                WorkflowLoader.DestroyObjectGraphFromDesignerHost(designerHost, designerHost.RootComponent as Activity);

            if (this.designSurface != null)
            {
                this.designSurface.Dispose();
                this.designSurface = null;
            }

            if (this.workflowView != null)
            {
                Controls.Remove(this.workflowView);
                this.workflowView.Dispose();
                this.workflowView = null;
            }
        }

        private void ShowDefaultWorkflow()
        {
            this.workflow = new SequentialWorkflowActivity();
            workflow.Name = "CustomWorkflow";

            this.LoadWorkflow();
        }

        private void LoadWorkflow()
        {
            using (StringWriter stringWriter = new StringWriter())
            {
                using (XmlWriter xmlWriter = XmlWriter.Create(stringWriter))
                {
                    WorkflowMarkupSerializer serializer = new WorkflowMarkupSerializer();
                    serializer.Serialize(xmlWriter, workflow);
                    this.Xoml = stringWriter.ToString();
                }
            }
        }

        public void ProcessZoom(int zoomFactor)
        {
            this.workflowView.Zoom = zoomFactor;
            this.workflowView.Update();
        }

        private void OnSelectionChanged(object sender, EventArgs e)
        {
            ISelectionService selectionService = GetService(typeof(ISelectionService)) as ISelectionService;

            if (selectionService != null)
            {
                this.propertyGrid.SelectedObjects = new ArrayList(selectionService.GetSelectedComponents()).ToArray();
            }
        }

        public void DeleteSelected()
        {
            ISelectionService selectionService = (ISelectionService)this.GetService(typeof(ISelectionService));

            if (selectionService != null)
            {
                if (selectionService.PrimarySelection is Activity)
                {
                    Activity activity = (Activity)selectionService.PrimarySelection;

                    if (activity.Name != this.WorkflowName)
                    {
                        activity.Parent.Activities.Remove(activity);
                        this.workflowView.Update();
                    }
                }
            }
        }

        public void LoadExistingWorkflow()
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "xoml files (*.xoml)|*.xoml|All files (*.*)|*.*";
            openFileDialog.FilterIndex = 1;
            openFileDialog.RestoreDirectory = true;

            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                using (XmlReader xmlReader = XmlReader.Create(openFileDialog.FileName))
                {
                    WorkflowMarkupSerializer serializer = new WorkflowMarkupSerializer();
                    this.workflow = (SequentialWorkflowActivity)serializer.Deserialize(xmlReader);
                    this.LoadWorkflow();

                    this.XomlFile = openFileDialog.FileName;
                    this.Text = "Designer Hosting Sample -- [" + openFileDialog.FileName + "]";
                }
            }
        }

        /// <summary>
        /// Saves an image of the System.Workflow.ComponentModel.Design.WorkflowView
        /// to the specified file using the specified image format.
        /// </summary>
        /// <param name="s">The System.IO.Stream to save the workflow view image to.</param>
        /// <param name="format">The System.Drawing.Imaging.ImageFormat to use to save the workflow view image.</param>
        public void SaveWorkflowImage(Stream s, ImageFormat format)
        {
            this.workflowView.SaveWorkflowImage(s, format);
        }

        /// <summary>
        /// Gets the RootActivity from the provided Child activity.
        /// </summary>
        /// <param name="child">The child.</param>
        /// <returns></returns>
        internal Activity GetRoot(Activity child)
        {
            Activity temp = child;
            while (temp.Parent != null)
            {
                temp = temp.Parent;
            }
            return temp;
        }

        /// <summary>
        /// Gets the parent activity for a StateMachineWorkflowActivity.
        /// </summary>
        /// <param name="child">The child.</param>
        /// <returns></returns>
        internal Activity[] GetStateParentActivity(Activity child)
        {
            if (child.Parent == null)
                return new Activity[] { null, null };
            Activity activity1 = child;
            Activity temp = null;
            while (activity1.GetType() != typeof(StateActivity) || activity1.Parent == null)
            {
                temp = activity1;
                activity1 = activity1.Parent;

            }
            return new Activity[] { activity1, temp };

        }

        public Type WFType
        {
            set
            {
                this.Xoml = LoadXomlForWFType(value);
            }
        }

        /// <summary>
        /// Highlights the provided activity.
        /// </summary>
        /// <param name="activityName">Name of the activity to highlight.</param>
        public void HighlightActivity(string activityName)
        {
            ISelectionService selectionService = (ISelectionService)designSurface.GetService(typeof(ISelectionService));
            IReferenceService referenceService = (IReferenceService)designSurface.GetService(typeof(IReferenceService));

            if (selectionService != null && referenceService != null)
            {
                Activity activityComponent = (Activity)referenceService.GetReference(activityName);

                if (activityComponent != null)
                {

                    IDesignerHost designerHost = designSurface.GetService(typeof(IDesignerHost)) as IDesignerHost;
                    if (designerHost != null)
                    {

                        Activity root = GetRoot(activityComponent);
                        bool statemachine = root is StateMachineWorkflowActivity;

                        if (statemachine)
                        {
                            Activity[] parents = GetStateParentActivity(activityComponent);
                            Activity state = parents[1];
                            Activity pstate = parents[0];
                            if (state != null && pstate != null)
                            {
                                ActivityDesigner sd = designerHost.GetDesigner(state) as ActivityDesigner;
                                FreeformActivityDesigner parentd = designerHost.GetDesigner(pstate) as FreeformActivityDesigner;
                                parentd.EnsureVisibleContainedDesigner(sd);
                            }
                        }

                        ActivityDesigner ad = designerHost.GetDesigner(activityComponent) as ActivityDesigner;
                        //ad.DesignerTheme.BackColorEnd = Color.OrangeRed;
                        //ad.DesignerTheme.BackColorStart = Color.OrangeRed;
                        selectionService.SetSelectedComponents(new IComponent[] { activityComponent as IComponent }, SelectionTypes.Primary);
                        workflowView.PerformLayout(true);
                    }
                }
            }
        }

        /// <summary>
        /// Loads the xoml for the workflow type.
        /// </summary>
        /// <param name="t">The type of workflow.</param>
        /// <returns>the xoml for the workflow type</returns>
        public string LoadXomlForWFType(Type t)
        {
            WorkflowMarkupSerializer xomlSerializer = new WorkflowMarkupSerializer();

            Activity a = (Activity)t.Assembly.CreateInstance(t.ToString());
            StringWriter writer = new StringWriter();
            XmlTextWriter xw = new XmlTextWriter(writer);
            xomlSerializer.Serialize(xw, a);
            return writer.ToString();
        }

        private void SaveFile()
        {
            if (this.XomlFile.Length != 0)
            {
                this.SaveExistingWorkflow(this.XomlFile);
            }
            else
            {
                SaveFileDialog saveFileDialog = new SaveFileDialog();
                saveFileDialog.Filter = "xoml files (*.xoml)|*.xoml|All files (*.*)|*.*";
                saveFileDialog.FilterIndex = 1;
                saveFileDialog.RestoreDirectory = true;

                if (saveFileDialog.ShowDialog() == DialogResult.OK)
                {
                    this.SaveExistingWorkflow(saveFileDialog.FileName);
                    this.Text = "Designer Hosting Sample -- [" + saveFileDialog.FileName + "]";
                }
            }
        }

        internal void SaveExistingWorkflow(string filePath)
        {
            if (this.designSurface != null && this.loader != null)
            {
                this.XomlFile = filePath;
                this.loader.PerformFlush();
            }
        }

        public bool Save()
        {
            return this.Save(true);
        }

        public bool Save(bool showMessage)
        {
            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.WaitCursor;

            bool saveOK = true;

            try
            {
                // Save the workflow first, and capture the filePath of the workflow
                this.SaveFile();

                XmlDocument doc = new XmlDocument();
                doc.Load(this.XomlFile);
                XmlAttribute attrib = doc.CreateAttribute("x", "Class", "http://schemas.microsoft.com/winfx/2006/xaml");
                attrib.Value = string.Format("{0}.{1}", this.GetType().Namespace, this.WorkflowName);
                doc.DocumentElement.Attributes.Append(attrib);
                doc.Save(this.XomlFile);

                if (showMessage)
                {
                    MessageBox.Show(this, "Workflow generated successfully. Generated xoml file:\n" + Path.Combine(Path.GetDirectoryName(this.GetType().Assembly.Location), this.XomlFile), this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                saveOK = false;
            }
            finally
            {
                this.Cursor = cursor;
            }

            return saveOK;
        }

        public bool Compile()
        {
            return this.Compile(true);
        }

        public bool Compile(bool showMessage)
        {
            if (!this.Save(false))
            {
                return false;
            }

            if (!File.Exists(this.XomlFile))
            {
                MessageBox.Show(this, "Cannot locate xoml file: " + Path.Combine(Path.GetDirectoryName(this.GetType().Assembly.Location), XomlFile), this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            bool compileOK = true;

            Cursor cursor = this.Cursor;
            this.Cursor = Cursors.WaitCursor;

            try
            {
                // Compile the workflow
                WorkflowCompiler compiler = new WorkflowCompiler();
                WorkflowCompilerParameters parameters = new WorkflowCompilerParameters();
                parameters.GenerateInMemory = true;
                WorkflowCompilerResults compilerResults = compiler.Compile(parameters, this.XomlFile);

                inMemoryAssembly = compilerResults.CompiledAssembly;

                StringBuilder errors = new StringBuilder();

                foreach (CompilerError compilerError in compilerResults.Errors)
                {
                    errors.Append(compilerError.ToString() + '\n');
                }

                if (errors.Length != 0)
                {
                    MessageBox.Show(this, errors.ToString(), this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    compileOK = false;
                }
                else if (showMessage)
                {
                    MessageBox.Show(this, "Workflow compiled successfully. Compiled assembly:\n" + compilerResults.CompiledAssembly.GetName(), this.Text, MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
            finally
            {
                this.Cursor = cursor;
            }

            return compileOK;
        }

        public bool Run()
        {
            if (inMemoryAssembly == null)
            {
                if (!this.Compile(false))
                {
                    return false;
                }
            }

            // Start the runtime engine
            if (this.workflowRuntime == null)
            {
                this.workflowRuntime = new WorkflowRuntime();
                this.workflowRuntime.StartRuntime();
            }

            this.workflowRuntime.WorkflowCompleted += new EventHandler<WorkflowCompletedEventArgs>(workflowRuntime_WorkflowCompleted);
            string typeName = string.Format("{0}.{1}", this.GetType().Namespace, this.WorkflowName);
            this.workflowRuntime.CreateWorkflow(AppDomain.CurrentDomain.CreateInstanceAndUnwrap(inMemoryAssembly.FullName, typeName).GetType()).Start();

            return true;
        }

        void workflowRuntime_WorkflowCompleted(object sender, WorkflowCompletedEventArgs e)
        {
            MessageBox.Show("Workflow complete");
        }

    }
}
